Broom = {}
Broom.ModDirectory = g_currentModDirectory
Broom.ModName = g_currentModName

local Broom_mt = Class(Broom, HandTool)
registerHandTool("Broom", Broom)

g_xmlManager:addInitSchemaFunction(function ()
	local schema = HandTool.xmlSchema

	schema:setXMLSpecializationType("Broom")
	schema:register(XMLValueType.NODE_INDEX, "handTool.Broom.data#node", "data node")
	schema:register(XMLValueType.NODE_INDEX, "handTool.Broom.data#ModelNode", "ModelNode")
	schema:register(XMLValueType.NODE_INDEX, "handTool.Broom.data#StateIdleTargetNode", "StateIdleTargetNode")
	schema:register(XMLValueType.NODE_INDEX, "handTool.Broom.data#StateActionTargetNode", "StateActionTargetNode")
	schema:register(XMLValueType.NODE_INDEX, "handTool.Broom.data#ClearAreaStart", "ClearAreaStart")
	schema:setXMLSpecializationType()
end)

function Broom:postLoad(xmlFile)
	if not Broom:superClass().postLoad(self, xmlFile) then
		return false
	end

	self.dataNode = xmlFile:getValue("handTool.Broom.data#node", nil, self.components, self.i3dMappings)
	self.ModelNode = xmlFile:getValue("handTool.Broom.data#ModelNode", nil, self.components, self.i3dMappings)

	self.StateIdleTargetNode = xmlFile:getValue("handTool.Broom.data#StateIdleTargetNode", nil, self.components, self.i3dMappings)
	self.StateActionTargetNode = xmlFile:getValue("handTool.Broom.data#StateActionTargetNode", nil, self.components, self.i3dMappings)

	self.ClearAreaStart = xmlFile:getValue("handTool.Broom.data#ClearAreaStart", nil, self.components, self.i3dMappings)
	self.ClearAreaWidth = getChildAt(self.ClearAreaStart, 0);
	self.ClearAreaHeight = getChildAt(self.ClearAreaStart, 1);

	self.StateIdleRaysNodes = {getChildAt(self.StateIdleTargetNode, 0), getChildAt(self.StateIdleTargetNode, 1), getChildAt(self.StateIdleTargetNode, 2)}
	self.StateActionRaysNodes = {getChildAt(self.StateActionTargetNode, 0), getChildAt(self.StateActionTargetNode, 1), getChildAt(self.StateActionTargetNode, 2)}

	self.SoundSample = g_soundManager:loadSampleFromXML(xmlFile, "handTool.Broom.sounds", "usage", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
	--g_soundManager:setSamplePitch(self.SoundSample, 0.01);
	--g_soundManager:setSampleVolumeOffset(self.SoundSample, 0.01);

	local sx, sy, sz = getScale(self.ModelNode)
	self.InitalModelScale = {x = sx, y = sy, z = sz}

	-- self.GizmoNodes = {}
	-- for i = 1, 8 do
	-- 	self.GizmoNodes[i] = loadI3DFile(Broom.ModDirectory .. "testcube.i3d")
	-- 	if self.GizmoNodes[i] then
	-- 		link(getRootNode(), self.GizmoNodes[i])
	-- 		setScale(self.GizmoNodes[i], 0.1, 0.1, 0.1)
	-- 	end
	-- end

	-- self.AnimationTargetNodeGizmo = self.GizmoNodes[7]
	-- self.ModelNodeGizmo = self.GizmoNodes[8]
	
	return true
end


function Broom.new(isServer, isClient, customMt)
	local self = Broom:superClass().new(isServer, isClient, customMt or Broom_mt)
	return self
end

function Broom:delete()
	
	-- for _, node in ipairs(self.GizmoNodes) do
	-- 	if node ~= nil then
	-- 		delete(node)
	-- 	end
	-- end
	g_soundManager:deleteSample(self.SoundSample)
	Broom:superClass().delete(self)
end

function Broom:onDeactivate()
	Broom:superClass().onDeactivate(self)
end

function Broom:GroundRayCallback(hitActorId, x, y, z, distance, nx, ny, nz, subShapeIndex, hitShapeId)
	self.RayResults[self.RayIndex] = {x, y, z, hitActorId == g_currentMission.terrainRootNode}
	return false
end

function Broom:GroundRaycasts(NodeList, dx, dy, dz)
	self.RayResults = {}
	dx, dy, dz = dx or 0, dy or -1, dz or 0

	for i = 1, #NodeList do
		self.RayIndex = i
		local node = NodeList[i]
		local x, y, z = getWorldTranslation(node)
		raycastAll(x, y, z, dx, dy, dz, "GroundRayCallback", 5, self, 12770)
	end

	if #self.RayResults == 0 then
		return (dy == -1) and self:GroundRaycasts(NodeList, 0, 1, 0) or nil
	end

	return self.RayResults;
end

function Broom:DrawDebugGizmos(IdleGroundRays, ActionGroundRays)

	-- local MergedGroundRays = {}
	-- for _, ray in ipairs(IdleGroundRays) do
	-- 	table.insert(MergedGroundRays, ray)
	-- end
	-- for _, ray in ipairs(ActionGroundRays) do
	-- 	table.insert(MergedGroundRays, ray)
	-- end

	-- for i = 1, #MergedGroundRays do
	-- 	--local x, y, z, terrain = MergedGroundRays[i]
	-- 	setWorldTranslation(self.GizmoNodes[i], MergedGroundRays[i][1], MergedGroundRays[i][2], MergedGroundRays[i][3])
	-- end
end

function Broom:RaysToPos(Rays, fallbackX, fallbackY, fallbackZ)
    if Rays == nil then
        return fallbackX or 0, fallbackY or 0, fallbackZ or 0
    end

    local validRays = {}
    for i = 1, 3 do
        if Rays[i] and Rays[i][1] and Rays[i][2] and Rays[i][3] then
            table.insert(validRays, Rays[i])
        end
    end

	if #validRays >= 3 then
		local totalY = validRays[1][2] + validRays[2][2] + validRays[3][2]
		local averageY = totalY / 3
		return validRays[1][1], averageY, validRays[1][3]
	end

	--fix max
	local maxY = -math.huge
	local x, z
	for _, ray in ipairs(validRays) do
		if ray[2] > maxY then
			maxY = ray[2]
			x = ray[1]
			z = ray[3]
		end
	end

	if x and z then
		return x, maxY, z
	else
		-- Return a default value if no valid rays are found
		return 0, 0, 0
	end
end

function Broom:updateAnimationTime(Active, AnimTime, dt)
	if self.AnimationTime == nil then
		self.AnimationTime = 0
	end

	if self.AnimationDirection == nil then
		self.AnimationDirection = 1
	end

	local animationSpeed = AnimTime -- Increased from 500 to make it slower

	if Active then
		self.AnimationTime = self.AnimationTime + (self.AnimationDirection * dt / animationSpeed)
		if self.AnimationTime >= 1 then
			self.AnimationTime = 1
			self.AnimationDirection = -1
		elseif self.AnimationTime <= 0 then
			self.AnimationTime = 0
			self.AnimationDirection = 1
		end
	else
		self.AnimationTime = math.max(self.AnimationTime - dt / animationSpeed, 0)
		self.AnimationDirection = 1
	end

	return self.AnimationTime;
end

function Broom:InterpolateValues(xCount, yMin, yMax)
    local t = math.max(0, math.min(1, xCount))
    local output = yMin + (yMax - yMin) * t
    return output
end

function Broom:ClearBroomAera()
	local xs, _, zs = getWorldTranslation(self.ClearAreaStart)
	local xw, _, zw = getWorldTranslation(self.ClearAreaWidth)
	local xh, _, zh = getWorldTranslation(self.ClearAreaHeight)

	DensityMapHeightUtil.clearArea(xs, zs, xw, zw, xh, zh)
end

function Broom:AudioController(Start)
	if Start then
		g_soundManager:playSample(self.SoundSample)
		else
		g_soundManager:stopSample(self.SoundSample)
	end
end

function Broom:update(dt, allowInput)

	local AnimationTime = self:updateAnimationTime(self:getIsInputActive(),750,dt)

	if AnimationTime > 0 then self:ClearBroomAera() end

	if AnimationTime > 0 then
		
		if not self.StartedLast then
			self.StartedLast = true;
			self:AudioController(true);
		end

		else

			if self.StartedLast then
				self.StartedLast = nil;
				self:AudioController(false);
			end
	end

	local fallbackX, fallbackY, fallbackZ = getWorldTranslation(self.StateIdleTargetNode)

	local IdleGroundRays = self:GroundRaycasts(self.StateIdleRaysNodes)
	local ActionGroundRays = self:GroundRaycasts(self.StateActionRaysNodes)
	
	local IdleX, IdleY, IdleZ = self:RaysToPos(IdleGroundRays, fallbackX, fallbackY, fallbackZ)
	local ActionX, ActionY, ActionZ = self:RaysToPos(ActionGroundRays, fallbackX, fallbackY, fallbackZ)

	if math.abs(IdleY - ActionY) > 2 then
		local clampedDifference = math.min(math.abs(IdleY - ActionY), 2)
		ActionY = IdleY + (ActionY > IdleY and clampedDifference or -clampedDifference)
	end

	local TargetAnimationOffsetX = self:InterpolateValues(AnimationTime, IdleX, ActionX)
	local TargetAnimationOffsetY = self:InterpolateValues(AnimationTime, IdleY, ActionY)
	local TargetAnimationOffsetZ = self:InterpolateValues(AnimationTime, IdleZ, ActionZ)

	local alpha = math.min(dt / 25, 1) 
	local currentX, currentY, currentZ = getWorldTranslation(self.ModelNode)
	local newX = currentX + (TargetAnimationOffsetX - currentX) * alpha
	local newY = currentY + (TargetAnimationOffsetY - currentY) * alpha
	local newZ = currentZ + (TargetAnimationOffsetZ - currentZ) * alpha

	setWorldTranslation(self.ModelNode, newX, newY, newZ)

	local _, cameraY, cameraZ = getWorldRotation(g_currentMission.player.cameraNode)
	local CameraX = self:InterpolateValues(AnimationTime, -0.30, -0.70)
	setWorldRotation(self.ModelNode, CameraX, cameraY + math.pi, cameraZ)


	local scaleY = cameraZ == 0 and self.InitalModelScale.y or -self.InitalModelScale.y
	local scaleZ = cameraZ == 0 and self.InitalModelScale.z or -self.InitalModelScale.z
	setScale(self.ModelNode, self.InitalModelScale.x, scaleY, scaleZ)

	--self:DrawDebugGizmos(IdleGroundRays, ActionGroundRays)

	Broom:superClass().update(self, dt)
	self:raiseActive()
end

function Broom:getIsInputActive()
	if self.activatePressed then
		self.activatePressed = false
		return true
	end

	return false
end

function Broom:getIsActiveForInput()
	if self.player == g_currentMission.player and not g_gui:getIsGuiVisible() then
		return true
	end

	return false
end

function Broom:isBeingUsed()
	return self:getIsInputActive()
end
